@@ -10,6 +10,7 @@
 #include "transport.h"
 #include "channel.h"
 #include <assert.h>
+#include "bootstrap.h"
 
 __thread int ncclGroupDepth = 0; // depth of ncclGroupStart nesting
 __thread ncclResult_t ncclGroupError = ncclSuccess;
@@ -22,7 +23,6 @@ __thread int ncclGroupBlocking = -1; /* default mode */
 __thread bool ncclGroupJobAbortFlag = false;
 
 void* ncclAsyncJobMain(void* arg);
-static ncclResult_t groupJobComplete(struct ncclGroupJob *job);
 
 ncclResult_t ncclAsyncLaunch(
     struct ncclAsyncJob* job,
@@ -32,6 +32,7 @@ ncclResult_t ncclAsyncLaunch(
   ) {
   ncclResult_t ret = ncclSuccess;
 
+  job->destroyFlag = comm->destroyFlag;
   if (ncclGroupDepth == 0) {
     ret = func(job);
     if (ret != ncclSuccess && undo) undo(job);
@@ -41,11 +42,15 @@ ncclResult_t ncclAsyncLaunch(
     job->undo = undo;
     job->destructor = destructor;
     job->abortFlag = comm->abortFlag;
+    job->abortFlagDev = comm->abortFlagDev;
     job->childAbortFlag = comm->childAbortFlag;
+    job->childAbortFlagDev = comm->childAbortFlagDev;
     job->state = ncclGroupJobRunning;
     job->comm = comm;
     /* check if there are blocking and nonblocking comms at the same time in group. */
-    if (ncclGroupBlocking == -1) {
+    if (comm->destroyFlag) {
+      ncclGroupBlocking = 1;
+    } else if (ncclGroupBlocking == -1) {
       /* first met communicator */
       ncclGroupBlocking = comm->config.blocking;
     } else if (ncclGroupBlocking != comm->config.blocking) {
@@ -99,11 +104,23 @@ exit:
   return ret;
 }
 
+NCCL_API(ncclResult_t, ncclGroupSimulateEnd, ncclSimInfo_t* simInfo);
+ncclResult_t ncclGroupSimulateEnd(ncclSimInfo_t* simInfo) {
+  ncclResult_t ret = ncclSuccess;
+  NVTX3_FUNC_RANGE_IN(nccl_domain);
+  NCCLCHECKGOTO(ncclGroupEndInternal(simInfo), ret, exit);
+  TRACE_CALL("ncclGroupSimulateEnd()");
+exit:
+  return ret;
+}
+
 struct ncclPreconnectJob {
   struct ncclAsyncJob base;
   struct ncclComm* comm;
+  bool* algoNeedConnect;
 };
-ncclResult_t ncclPreconnectFunc(struct ncclAsyncJob* job_) {
+
+ncclResult_t ncclP2PPreconnectFunc(struct ncclAsyncJob* job_) {
   struct ncclPreconnectJob* job = (struct ncclPreconnectJob*)job_;
   struct ncclComm* comm = job->comm;
   CUDACHECK(cudaSetDevice(comm->cudaDev));
@@ -112,6 +129,57 @@ ncclResult_t ncclPreconnectFunc(struct ncclAsyncJob* job_) {
   return ncclSuccess;
 }
 
+ncclResult_t ncclCollPreconnectFunc(struct ncclAsyncJob* job_) {
+  struct ncclPreconnectJob* job = (struct ncclPreconnectJob*)job_;
+  struct ncclComm* comm = job->comm;
+  ncclResult_t ret = ncclSuccess;
+
+  CUDACHECK(cudaSetDevice(comm->cudaDev));
+  if (CPU_COUNT(&comm->cpuAffinity)) sched_setaffinity(0, sizeof(cpu_set_t), &comm->cpuAffinity);
+  for (int i = 0; i < NCCL_NUM_ALGORITHMS; ++i) {
+    if (job->algoNeedConnect[i]) {
+      switch (i) {
+        case NCCL_ALGO_RING: {
+          NCCLCHECKGOTO(ncclTransportRingConnect(comm), ret, fail);
+          break;
+        }
+        case NCCL_ALGO_TREE: {
+          NCCLCHECKGOTO(ncclTransportTreeConnect(comm), ret, fail);
+          break;
+        }
+        case NCCL_ALGO_NVLS: {
+          /* If we are using NVLS_TREE algo, we must mark NVLS algo to set up
+           * NVLS intra-node buffer */
+          NCCLCHECKGOTO(ncclNvlsBufferSetup(comm), ret, fail);
+          break;
+        }
+        case NCCL_ALGO_NVLS_TREE: {
+          NCCLCHECKGOTO(ncclNvlsTreeConnect(comm), ret, fail);
+          break;
+        }
+        case NCCL_ALGO_COLLNET_CHAIN: {
+          NCCLCHECKGOTO(ncclCollNetChainBufferSetup(comm), ret, fail);
+          break;
+        }
+        case NCCL_ALGO_COLLNET_DIRECT: {
+          NCCLCHECKGOTO(ncclCollNetDirectBufferSetup(comm), ret, fail);
+          break;
+        }
+        default: {
+          ret = ncclInternalError;
+          goto fail;
+        }
+      }
+    }
+  }
+
+exit:
+  free(job->algoNeedConnect);
+  return ret;
+fail:
+  goto exit;
+}
+
 static ncclResult_t doLaunches(struct ncclComm* head) {
   ncclResult_t result = ncclSuccess;
   struct ncclComm* cliqueComm0 = head->intraComm0;
@@ -125,7 +193,7 @@ static ncclResult_t doLaunches(struct ncclComm* head) {
     struct ncclComm* comm = cliqueHead;
     bool capturingYes = false, capturingNo = false;
     do {
-      (ncclCudaGraphValid(comm->tasks.capturingGraph) ? capturingYes : capturingNo) = true;
+      (ncclCudaGraphValid(comm->planner.capturingGraph) ? capturingYes : capturingNo) = true;
       CUDACHECKGOTO(cudaSetDevice(comm->cudaDev), result, failure);
       NCCLCHECKGOTO(ncclLaunchPrepare(comm), result, failure);
       if (useBarrier) ncclCommIntraBarrierIn(comm, 1);
@@ -143,7 +211,7 @@ static ncclResult_t doLaunches(struct ncclComm* head) {
     }
 
     while (true) { // Iterate rounds of launches for clique.
-      bool moreRounds;
+      bool moreRounds = false;
       comm = cliqueHead;
       do { // Iterate clique members.
         struct ncclComm* next = comm->groupNext;
@@ -151,19 +219,19 @@ static ncclResult_t doLaunches(struct ncclComm* head) {
           // Barrier reduction result tells us if this was the final round.
           moreRounds = 0 != ncclCommIntraBarrierOut(comm);
         } else {
-          moreRounds = comm->unlaunchedPlansHead != nullptr;
+          moreRounds |= comm->planner.unlaunchedPlansHead != nullptr;
         }
         if (moreRounds) {
           // Pop next unlaunched kernel
-          struct ncclKernelPlan* plan = comm->unlaunchedPlansHead;
+          struct ncclKernelPlan* plan = comm->planner.unlaunchedPlansHead;
           if (plan != nullptr) {
-            comm->unlaunchedPlansHead = plan->next;
+            comm->planner.unlaunchedPlansHead = plan->next;
             CUDACHECKGOTO(cudaSetDevice(comm->cudaDev), result, failure);
             NCCLCHECKGOTO(ncclLaunchKernelBefore_NoUncapturedCuda(comm, plan), result, failure);
             NCCLCHECKGOTO(ncclLaunchKernel(comm, plan), result, failure);
           }
           // Barrier reduction input indicates if we require further rounds.
-          if (useBarrier) ncclCommIntraBarrierIn(comm, comm->unlaunchedPlansHead != nullptr ? 1 : 0);
+          if (useBarrier) ncclCommIntraBarrierIn(comm, comm->planner.unlaunchedPlansHead != nullptr ? 1 : 0);
           if (plan != nullptr) {
             NCCLCHECKGOTO(ncclLaunchKernelAfter_NoCuda(comm, plan), result, failure);
           }
@@ -181,9 +249,28 @@ failure:
   return result;
 }
 
-static void groupCleanup(struct ncclComm** groupCommHeadPtr, struct ncclComm** groupCommPreconnectHeadPtr, struct ncclIntruQueue<struct ncclAsyncJob, &ncclAsyncJob::next>* asyncJobsPtr, ncclResult_t* groupErrorPtr, ncclResult_t error) {
+static inline void groupResetJobState(struct ncclGroupJob* job) {
+  if (job) {
+    if (job->groupBlockingPtr) *job->groupBlockingPtr = -1;
+    if (job->abortFlagPtr) *job->abortFlagPtr = false;
+    if (job->groupErrorPtr) *job->groupErrorPtr = ncclSuccess;
+    if (job->groupCommHeadPtr) *job->groupCommHeadPtr = NULL;
+    if (job->groupCommPreconnectHeadPtr) *job->groupCommPreconnectHeadPtr = NULL;
+    memset(job, 0, sizeof(struct ncclGroupJob));
+  }
+  return;
+}
+
+static void groupCleanup(struct ncclComm** groupCommHeadPtr, struct ncclComm** groupCommPreconnectHeadPtr, struct ncclIntruQueue<struct ncclAsyncJob, &ncclAsyncJob::next>* asyncJobsPtr, ncclResult_t* groupErrorPtr, int* groupBlockingPtr, volatile bool* groupJobAbortFlagPtr, ncclResult_t error) {
   struct ncclComm* comm = *groupCommHeadPtr;
 
+  /* reset all thread local variables */
+  *groupCommHeadPtr = NULL;
+  *groupCommPreconnectHeadPtr = NULL;
+  *groupErrorPtr = ncclSuccess;
+  *groupBlockingPtr = -1;
+  *groupJobAbortFlagPtr = false;
+
   while (comm != nullptr) {
     struct ncclComm* next = comm->groupNext;
     (void) ncclGroupCommLeave(comm); // overwrites comm->groupNext
@@ -192,37 +279,29 @@ static void groupCleanup(struct ncclComm** groupCommHeadPtr, struct ncclComm** g
     // is needed.
     comm->preconnectNext = reinterpret_cast<struct ncclComm*>(0x1);
     for (int i = 0; i < comm->nRanks; i++) {
-      comm->tasks.peers[i].sendSeen = false;
-      comm->tasks.peers[i].recvSeen = false;
       comm->connectSend[i] = 0UL;
       comm->connectRecv[i] = 0UL;
     }
-    comm->unlaunchedPlansHead = nullptr;
     // Reclaim abandoned kernel plan memory. Note ncclWork structs were already
     // reclaimed by a `ncclMemoryStackPop(&comm->memScoped)` during `ncclGroupCommLeave()`.
-    while (!ncclIntruQueueEmpty(&comm->planQueue)) {
-      struct ncclKernelPlan* plan = ncclIntruQueueDequeue(&comm->planQueue);
+    while (!ncclIntruQueueEmpty(&comm->planner.planQueue)) {
+      struct ncclKernelPlan* plan = ncclIntruQueueDequeue(&comm->planner.planQueue);
       // Persistent plans will be reclaimed via the callbackQueue when the
       // graph drops its UserObject reference.
       if (!plan->persistent) {
-        for (int c = 0; c < MAXCHANNELS; c++) {
-          while (!ncclIntruQueueEmpty(&plan->channels[c].proxyOpQueue)) {
-            struct ncclProxyOp* pxop = ncclIntruQueueDequeue(&plan->channels[c].proxyOpQueue);
-            ncclMemoryPoolFree(&comm->memPool_ncclProxyOp, pxop);
-          }
+        while (!ncclIntruQueueEmpty(&plan->proxyOpQueue)) {
+          struct ncclProxyOp* pxop = ncclIntruQueueDequeue(&plan->proxyOpQueue);
+          ncclMemoryPoolFree(&comm->memPool_ncclProxyOp, pxop);
         }
         ncclMemoryPoolFree(&comm->memPool_ncclKernelPlan, plan);
       }
     }
-    // Reset comm->tasks to empty.
-    comm->tasks.nTasksColl = 0;
-    comm->tasks.nTasksP2p = 0;
-    comm->tasks.streams = nullptr;
-    ncclIntruQueueConstruct(&comm->tasks.collQueue);
-    comm->tasks.collBytesTotal = 0;
-    for (int i = 0; i < comm->nRanks; i++) {
-      ncclIntruQueueConstruct(&comm->tasks.peers[i].sendQueue);
-      ncclIntruQueueConstruct(&comm->tasks.peers[i].recvQueue);
+
+    { // Reset comm->planner to empty.
+      ncclKernelPlanner::Peer* tmp = comm->planner.peers;
+      memset(&comm->planner, 0, sizeof(comm->planner));
+      comm->planner.peers = tmp;
+      memset(comm->planner.peers, 0, comm->nRanks*sizeof(comm->planner.peers[0]));
     }
 
     if (!comm->config.blocking)
@@ -233,50 +312,19 @@ static void groupCleanup(struct ncclComm** groupCommHeadPtr, struct ncclComm** g
   /* reset everything */
   while (!ncclIntruQueueEmpty(asyncJobsPtr)) {
     struct ncclAsyncJob* job = ncclIntruQueueDequeue(asyncJobsPtr);
-    *job->abortFlag = 1;
     if (job->comm && !job->comm->config.blocking)
       (void) ncclCommSetAsyncError(job->comm, error);
     if (job->undo) job->undo(job);
     if (job->destructor) job->destructor((void*)job);
   }
 
-  *groupErrorPtr = ncclSuccess;
-  *groupCommHeadPtr = nullptr;
-  *groupCommPreconnectHeadPtr = nullptr;
   return;
 }
 
-static ncclResult_t groupLaunch(struct ncclAsyncJob *job_) {
-  int savedDev;
+static ncclResult_t asyncJobLaunch(struct ncclIntruQueue<struct ncclAsyncJob, &ncclAsyncJob::next> *asyncJobsMain, volatile bool *groupAbortFlag) {
   ncclResult_t ret = ncclSuccess;
   bool jobsDone = false;
   bool errorJobAbortFlag = false;
-  struct ncclGroupJob *gjob = (struct ncclGroupJob*) job_;
-  struct ncclComm *groupCommHeadMain = *gjob->groupCommHeadPtr;
-  struct ncclComm *groupCommPreconnectHeadMain = *gjob->groupCommPreconnectHeadPtr;
-  struct ncclIntruQueue<struct ncclAsyncJob, &ncclAsyncJob::next> *asyncJobsMain = gjob->asyncJobsPtr;
-  volatile bool *groupAbortFlag = gjob->abortFlagPtr;
-
-  CUDACHECKGOTO(cudaGetDevice(&savedDev), ret, fail);
-
-  if (groupCommPreconnectHeadMain != nullptr) {
-    struct ncclComm* comm = groupCommPreconnectHeadMain;
-    do {
-      struct ncclPreconnectJob* job;
-      NCCLCHECKGOTO(ncclCalloc(&job, 1), ret, fail);
-      job->base.func = ncclPreconnectFunc;
-      job->base.undo = nullptr;
-      job->base.destructor = free;
-      job->base.state = ncclGroupJobRunning;
-      job->base.abortFlag = comm->abortFlag;
-      job->comm = comm;
-      ncclIntruQueueEnqueue(asyncJobsMain, &job->base);
-
-      struct ncclComm* next = comm->preconnectNext;
-      comm->preconnectNext = reinterpret_cast<struct ncclComm*>(0x1);
-      comm = next;
-    } while (comm != nullptr);
-  }
 
   if (!ncclIntruQueueEmpty(asyncJobsMain)) {
     struct ncclAsyncJob* job = ncclIntruQueueHead(asyncJobsMain);
@@ -307,9 +355,13 @@ static ncclResult_t groupLaunch(struct ncclAsyncJob *job_) {
           assert(state == ncclGroupJobJoined);
         }
 
-        if (*groupAbortFlag == true || errorJobAbortFlag == true) {
-          *job->abortFlag = 1;
-          if (job->childAbortFlag) *job->childAbortFlag = 1;
+        if (!job->destroyFlag && (__atomic_load_n(groupAbortFlag, __ATOMIC_ACQUIRE) || errorJobAbortFlag == true)) {
+          __atomic_store_n(job->abortFlag, 1, __ATOMIC_RELEASE);
+          __atomic_store_n(job->abortFlagDev, 1, __ATOMIC_RELEASE);
+          if (job->childAbortFlag) {
+            __atomic_store_n(job->childAbortFlag, 1, __ATOMIC_RELEASE);
+            __atomic_store_n(job->childAbortFlagDev, 1, __ATOMIC_RELEASE);
+          }
         }
 
         job = job->next;
@@ -321,20 +373,86 @@ static ncclResult_t groupLaunch(struct ncclAsyncJob *job_) {
     if (ret != ncclSuccess) goto fail;
   }
 
-  if (groupCommHeadMain != nullptr) {
-    NCCLCHECKGOTO(doLaunches(groupCommHeadMain), ret, fail);
-  }
-
-  /* this atomic must happen before cleanup and setting state of communicators */
-  __atomic_store_n(&gjob->doneFlag, true, __ATOMIC_RELEASE);
-
   while (!ncclIntruQueueEmpty(asyncJobsMain)) {
     struct ncclAsyncJob* job = ncclIntruQueueDequeue(asyncJobsMain);
-    if (job->comm && !job->comm->config.blocking)
+    if (!job->destroyFlag && job->comm && !job->comm->config.blocking)
       (void) ncclCommSetAsyncError(job->comm, ret);
     if (job->destructor) job->destructor((void*)job);
   }
 
+exit:
+  return ret;
+fail:
+  goto exit;
+}
+
+static ncclResult_t groupLaunch(struct ncclAsyncJob *job_, ncclSimInfo_t* simInfo = NULL) {
+  int savedDev;
+  ncclResult_t ret = ncclSuccess;
+  struct ncclGroupJob *gjob = (struct ncclGroupJob*) job_;
+  struct ncclComm *groupCommHeadMain = *gjob->groupCommHeadPtr;
+  struct ncclComm *groupCommPreconnectHeadMain = *gjob->groupCommPreconnectHeadPtr;
+  struct ncclIntruQueue<struct ncclAsyncJob, &ncclAsyncJob::next> *asyncJobsMain = gjob->asyncJobsPtr;
+  bool *groupAbortFlag = gjob->abortFlagPtr;
+
+  CUDACHECKGOTO(cudaGetDevice(&savedDev), ret, fail);
+
+  if (!simInfo && groupCommPreconnectHeadMain != nullptr) {
+    struct ncclComm* comm = groupCommPreconnectHeadMain;
+    do {
+      struct ncclPreconnectJob* job;
+      NCCLCHECKGOTO(ncclCalloc(&job, 1), ret, fail);
+      job->base.func = ncclP2PPreconnectFunc;
+      job->base.undo = nullptr;
+      job->base.destructor = free;
+      job->base.state = ncclGroupJobRunning;
+      job->base.abortFlag = comm->abortFlag;
+      job->base.abortFlagDev = comm->abortFlagDev;
+      job->comm = comm;
+      ncclIntruQueueEnqueue(asyncJobsMain, &job->base);
+
+      struct ncclComm* next = comm->preconnectNext;
+      comm->preconnectNext = reinterpret_cast<struct ncclComm*>(0x1);
+      comm = next;
+    } while (comm != nullptr);
+  }
+
+  NCCLCHECKGOTO(asyncJobLaunch(asyncJobsMain, groupAbortFlag), ret, fail);
+
+  /* Connect channels at runtime if cumem is supported */
+  if (groupCommHeadMain != nullptr) {
+    struct ncclComm* comm = groupCommHeadMain;
+
+    do {
+      bool needConnect = false;
+      bool algoNeedConnect[NCCL_NUM_ALGORITHMS];
+      memset(algoNeedConnect, 0, sizeof(bool) * NCCL_NUM_ALGORITHMS);
+
+      NCCLCHECKGOTO(ncclPrepareTasks(comm, algoNeedConnect, &needConnect, simInfo), ret, fail);
+
+      if (comm->cuMemSupport && needConnect) {
+        struct ncclPreconnectJob* job;
+        NCCLCHECKGOTO(ncclCalloc(&job, 1), ret, fail);
+        job->base.func = ncclCollPreconnectFunc;
+        job->base.undo = nullptr;
+        job->base.destructor = free;
+        job->base.state = ncclGroupJobRunning;
+        job->base.abortFlag = comm->abortFlag;
+        job->comm = comm;
+        NCCLCHECKGOTO(ncclCalloc(&job->algoNeedConnect, NCCL_NUM_ALGORITHMS), ret, fail);
+        memcpy(job->algoNeedConnect, algoNeedConnect, sizeof(bool) * NCCL_NUM_ALGORITHMS);
+        ncclIntruQueueEnqueue(asyncJobsMain, &job->base);
+      }
+      comm = comm->groupNext;
+    } while (comm);
+
+    NCCLCHECKGOTO(asyncJobLaunch(asyncJobsMain, groupAbortFlag), ret, fail);
+  }
+
+  if ((!simInfo) && (groupCommHeadMain != nullptr)) {
+    NCCLCHECKGOTO(doLaunches(groupCommHeadMain), ret, fail);
+  }
+
   while (groupCommHeadMain != nullptr) {
     struct ncclComm* comm = groupCommHeadMain;
     struct ncclComm* next = comm->groupNext;
@@ -345,21 +463,26 @@ static ncclResult_t groupLaunch(struct ncclAsyncJob *job_) {
     groupCommHeadMain = next;
   }
 
-  *gjob->groupErrorPtr = ncclSuccess;
-  *gjob->groupCommHeadPtr = nullptr;
-  *gjob->groupCommPreconnectHeadPtr = nullptr;
-
   CUDACHECK(cudaSetDevice(savedDev));
 
 exit:
   return ret;
 fail:
-  groupCleanup(gjob->groupCommHeadPtr, gjob->groupCommPreconnectHeadPtr, gjob->asyncJobsPtr, gjob->groupErrorPtr, ret);
+  groupCleanup(gjob->groupCommHeadPtr, gjob->groupCommPreconnectHeadPtr, gjob->asyncJobsPtr, gjob->groupErrorPtr, gjob->groupBlockingPtr, gjob->abortFlagPtr, ret);
   goto exit;
 }
 
-ncclResult_t ncclGroupEndInternal() {
+static ncclResult_t groupLaunchNonBlocking(struct ncclAsyncJob *job_) {
+  return groupLaunch(job_ /* estimatedTime = NULL */);
+}
+
+ncclResult_t ncclGroupEndInternal(ncclSimInfo_t* simInfo) {
   ncclResult_t ret = ncclSuccess;
+  ncclSimInfo_t internalSimInfo = NCCL_SIM_INFO_INITIALIZER;
+  ncclSimInfo_t* internalSimInfoPtr = NULL;
+  size_t realSize = 0;
+
+  internalSimInfo.magic = 0;
 
   if (ncclGroupDepth == 0) {
     WARN("ncclGroupEnd: not in a group call.");
@@ -371,13 +494,26 @@ ncclResult_t ncclGroupEndInternal() {
 
   if ((ret = ncclGroupError) != ncclSuccess) goto fail;
 
+  if (simInfo) {
+    memcpy((void*)&realSize, (void*)&simInfo->size, sizeof(size_t));
+    realSize = realSize > sizeof(ncclSimInfo_t) ? sizeof(ncclSimInfo_t) : realSize;
+    memcpy((void*)&internalSimInfo, (void*)simInfo, realSize);
+    if (internalSimInfo.magic != 0x74685283) {
+      WARN("ncclSimInfo_t argument not initialized via NCCL_SIM_INFO_INITIALIZER");
+      ret = ncclInvalidArgument;
+      goto fail;
+    }
+    internalSimInfoPtr = &internalSimInfo;
+  }
+
   if (ncclGroupCommHead != nullptr || !ncclIntruQueueEmpty(&ncclAsyncJobs) || ncclGroupCommPreconnectHead != nullptr) {
     ncclGroupJobMain.groupCommHeadPtr = &ncclGroupCommHead;
     ncclGroupJobMain.groupCommPreconnectHeadPtr = &ncclGroupCommPreconnectHead;
     ncclGroupJobMain.groupErrorPtr = &ncclGroupError;
     ncclGroupJobMain.asyncJobsPtr = &ncclAsyncJobs;
     ncclGroupJobMain.abortFlagPtr = &ncclGroupJobAbortFlag;
-    ncclGroupJobMain.doneFlag = false;
+    ncclGroupJobMain.groupBlockingPtr = &ncclGroupBlocking;
+    ncclGroupJobMain.initialized = true;
     ncclGroupJobMainPtr = &ncclGroupJobMain;
     /* make sure ncclGroupBlocking has been set. */
     assert(ncclGroupBlocking == 0 || ncclGroupBlocking == 1);
@@ -387,6 +523,7 @@ ncclResult_t ncclGroupEndInternal() {
         ncclAsyncJob* job = ncclIntruQueueHead(&ncclAsyncJobs);
         do {
           NCCLCHECKGOTO(ncclCommSetAsyncError(job->comm, ncclInProgress), ret, fail);
+          job->comm->groupJob = ncclGroupJobMainPtr;
           job = job->next;
         } while (job);
       }
@@ -395,30 +532,43 @@ ncclResult_t ncclGroupEndInternal() {
         ncclComm_t comm = ncclGroupCommHead;
         do {
           NCCLCHECKGOTO(ncclCommSetAsyncError(comm, ncclInProgress), ret, fail);
+          /* link group job to communicators. */
+          comm->groupJob = ncclGroupJobMainPtr;
           comm = comm->groupNext;
         } while (comm);
       }
-      ncclGroupJobMainPtr->base.func = groupLaunch;
+
+      ncclGroupJobMainPtr->base.func = groupLaunchNonBlocking;
       SYSCHECKGOTO(pthread_create(&ncclGroupJobMainPtr->base.thread, NULL, ncclAsyncJobMain, (void*)&ncclGroupJobMainPtr->base), ret, fail);
       ret = ncclInProgress;
     } else {
       /* blocking group */
-      NCCLCHECKGOTO(groupLaunch(&ncclGroupJobMainPtr->base), ret, fail);
-      groupResetJobState();
+      NCCLCHECKGOTO(groupLaunch(&ncclGroupJobMainPtr->base, internalSimInfoPtr), ret, fail);
+      if (simInfo) memcpy((void*)simInfo, (void*)internalSimInfoPtr, realSize);
+      groupResetJobState(ncclGroupJobMainPtr);
     }
   }
 
 exit:
   return ret;
 fail:
-  groupCleanup(&ncclGroupCommHead, &ncclGroupCommPreconnectHead, &ncclAsyncJobs, &ncclGroupError, ret);
-  groupResetJobState();
+  groupCleanup(&ncclGroupCommHead, &ncclGroupCommPreconnectHead, &ncclAsyncJobs, &ncclGroupError, &ncclGroupBlocking, &ncclGroupJobAbortFlag, ret);
   goto exit;
 }
 
-void ncclGroupJobAbort() {
-  ncclGroupJobAbortFlag = true;
-  (void) groupJobComplete(ncclGroupJobMainPtr);
-  /* reset group abort flag */
-  ncclGroupJobAbortFlag = false;
+ncclResult_t ncclGroupJobComplete(struct ncclGroupJob* groupJob) {
+  ncclResult_t ret = ncclSuccess;
+  if (groupJob && groupJob->initialized) {
+    ret = ncclAsyncJobComplete(&groupJob->base);
+    groupResetJobState(groupJob);
+  }
+  return ret;
+}
+
+ncclResult_t ncclGroupJobAbort(struct ncclGroupJob* groupJob) {
+  if (groupJob && groupJob->initialized) {
+    __atomic_store_n(groupJob->abortFlagPtr, true, __ATOMIC_RELEASE);
+    NCCLCHECK(ncclGroupJobComplete(groupJob));
+  }
+  return ncclSuccess;
 }
