@@ -6,6 +6,7 @@
 
 #include "comm.h"
 #include "shm.h"
+#include "transport.h"
 
 struct shmConnectInfo {
   char shmName[7];
@@ -150,9 +151,10 @@ static ncclResult_t shmSendConnect(struct ncclComm* comm, struct ncclConnect* co
   }
   send->conn.tail = &resources->devRemHostMem->tail;
   send->conn.head = &resources->devHostMem->head;
+  send->conn.stepSize = comm->buffSizes[NCCL_PROTO_SIMPLE]/NCCL_STEPS;
 
   if (useMemcpyRecv) {
-    send->conn.sizesFifo = resources->devRemHostMem->sizesFifo;
+    send->conn.connFifo = resources->devRemHostMem->connFifo;
   }
   if (useMemcpySend) {
     int tpProxyRank;
@@ -162,8 +164,12 @@ static ncclResult_t shmSendConnect(struct ncclComm* comm, struct ncclConnect* co
     NCCLCHECK(ncclProxyCallBlocking(comm, &send->proxyConn, ncclProxyMsgConnect, &proxyInfo, sizeof(struct shmProxyInfo), &proxyInfo, sizeof(struct shmProxyInfo)));
     send->conn.buffs[NCCL_PROTO_SIMPLE] = proxyInfo.devFifo;
     send->conn.tail = &proxyInfo.ceRecvMem->tail;
-    send->conn.sizesFifo = proxyInfo.ceRecvMem->sizesFifo;
+    send->conn.connFifo = proxyInfo.ceRecvMem->connFifo;
   }
+
+  // We must assign the proxyConn's proxyProgress property for proper checking at enqueue-time
+  send->proxyConn.proxyProgress = shmTransport.send.proxyProgress;
+
   return ncclSuccess;
 }
 
@@ -185,6 +191,7 @@ static ncclResult_t shmRecvConnect(struct ncclComm* comm, struct ncclConnect* co
   }
   recv->conn.head = &resources->devRemHostMem->head;
   recv->conn.tail = &resources->devHostMem->tail;
+  recv->conn.stepSize = comm->buffSizes[NCCL_PROTO_SIMPLE]/NCCL_STEPS;
 
   if (useMemcpyRecv) {
     NCCLCHECK(ncclProxyConnect(comm, TRANSPORT_SHM, 0, comm->rank, &recv->proxyConn));
@@ -193,6 +200,10 @@ static ncclResult_t shmRecvConnect(struct ncclComm* comm, struct ncclConnect* co
     recv->conn.buffs[NCCL_PROTO_SIMPLE] = proxyInfo.devFifo;
     recv->conn.tail = &proxyInfo.ceRecvMem->tail;
   }
+
+  // We must assign the proxyConn's proxyProgress property for proper checking at enqueue-time
+  recv->proxyConn.proxyProgress = shmTransport.recv.proxyProgress;
+
   return ncclSuccess;
 }
 
@@ -202,6 +213,7 @@ static ncclResult_t shmSendFree(struct ncclConnector* send) {
     NCCLCHECK(ncclShmClose(resources->hostHandle));
     NCCLCHECK(ncclShmClose(resources->remHandle));
     free(resources);
+    send->transportResources = NULL;
   }
   return ncclSuccess;
 }
@@ -212,6 +224,7 @@ static ncclResult_t shmRecvFree(struct ncclConnector* recv) {
     NCCLCHECK(ncclShmClose(resources->hostHandle));
     NCCLCHECK(ncclShmClose(resources->remHandle));
     free(resources);
+    recv->transportResources = NULL;
   }
   return ncclSuccess;
 }
@@ -263,6 +276,7 @@ static ncclResult_t shmSendProxyFree(struct ncclProxyConnection* connection, str
       CUDACHECK(cudaEventDestroy(resources->events[i]));
     }
     free(connection->transportResources);
+    connection->transportResources = NULL;
   }
   return ncclSuccess;
 }
@@ -278,6 +292,7 @@ static ncclResult_t shmRecvProxyFree(struct ncclProxyConnection* connection, str
       CUDACHECK(cudaEventDestroy(resources->events[i]));
     }
     free(connection->transportResources);
+    connection->transportResources = NULL;
   }
   return ncclSuccess;
 }
@@ -307,15 +322,15 @@ static ncclResult_t shmSendProxyProgress(struct ncclProxyState* proxyState, stru
       }
       if (sub->transmitted < sub->done + NCCL_STEPS && sub->transmitted < sub->nsteps) {
         int buffSlot = (sub->base+sub->transmitted)%NCCL_STEPS;
-        volatile int* sizesFifo = resources->ceRecvMem->sizesFifo;
+        volatile struct ncclConnFifo* connFifo = resources->ceRecvMem->connFifo;
         volatile uint64_t* recvTail = &resources->ceRecvMem->tail;
         // Check GPU has sent everything
         if ((*recvTail > sub->base+sub->transmitted)) {
-          int size = sizesFifo[buffSlot];
+          int size = connFifo[buffSlot].size;
           CUDACHECK(cudaMemcpyAsync(resources->shmFifo+buffSlot*stepSize, resources->devFifo+buffSlot*stepSize, size, cudaMemcpyDeviceToHost, resources->stream));
           CUDACHECK(cudaEventRecord(resources->events[buffSlot], resources->stream));
-          resources->recvMem->sizesFifo[buffSlot] = size;
-          __sync_synchronize(); // make sure sizesFifo is visible
+          resources->recvMem->connFifo[buffSlot].size = size;
+          __sync_synchronize(); // make sure connFifo[].size is visible
           sub->transmitted += args->sliceSteps;
         }
       }
@@ -366,11 +381,11 @@ static ncclResult_t shmRecvProxyProgress(struct ncclProxyState* proxyState, stru
       }
       if (sub->transmitted < sub->done + NCCL_STEPS && sub->transmitted < sub->nsteps) {
         int buffSlot = (sub->base+sub->transmitted)%NCCL_STEPS;
-        volatile int* sizesFifo = resources->recvMem->sizesFifo;
+        volatile struct ncclConnFifo* connFifo = resources->recvMem->connFifo;
         volatile uint64_t* recvTail = &resources->recvMem->tail;
         // Check data is ready in SHM
         if ((*recvTail > sub->base+sub->transmitted)) {
-          int size = sizesFifo[buffSlot];
+          int size = connFifo[buffSlot].size;
           CUDACHECK(cudaMemcpyAsync(resources->devFifo+buffSlot*stepSize, resources->shmFifo+buffSlot*stepSize, size, cudaMemcpyHostToDevice, resources->stream));
           CUDACHECK(cudaEventRecord(resources->events[buffSlot], resources->stream));
           sub->transmitted += args->sliceSteps;
@@ -401,8 +416,8 @@ static ncclResult_t shmRecvProxyProgress(struct ncclProxyState* proxyState, stru
 struct ncclTransport shmTransport = {
   "SHM",
   shmCanConnect,
-  { shmSendSetup, shmSendConnect, shmSendFree, NULL, NULL, NULL, NULL, NULL },
-  { shmRecvSetup, shmRecvConnect, shmRecvFree, NULL, NULL, NULL, NULL, NULL }
+  { shmSendSetup, shmSendConnect, shmSendFree, NULL, NULL, NULL, NULL, NULL, NULL },
+  { shmRecvSetup, shmRecvConnect, shmRecvFree, NULL, NULL, NULL, NULL, NULL, NULL }
 };
 
 static void initCeOperation() {
